<?php

declare(strict_types=1);

if (!defined('_PS_VERSION_')) {
    exit;
}

include_once(_PS_MODULE_DIR_ . 'conotoxiapay/tools/sdk/internals/autoload.php');
include_once(_PS_MODULE_DIR_ . 'conotoxiapay/tools/class-conotoxia-pay-storage.php');

use CKPL\Pay\Configuration\ConfigurationInterface;
use CKPL\Pay\Configuration\Factory\ConfigurationFactory;
use CKPL\Pay\Pay;

use PrestaShop\Module\ConotoxiaPay\Definitions;
use PrestaShop\Module\ConotoxiaPay\Install\InstallerFactory;
use PrestaShop\Module\ConotoxiaPay\Repository\PaymentRepository;
use PrestaShop\Module\ConotoxiaPay\Repository\RefundRepository;

use PrestaShop\PrestaShop\Core\Payment\PaymentOption;

require_once __DIR__ . '/vendor/autoload.php';

class ConotoxiaPay extends PaymentModule
{

    /**
     * @type string
     */
    const PAYMENTS_HOST = 'https://partner.cinkciarz.pl';

    /**
     * @type string
     */
    const OIDC_HOST = 'https://login.cinkciarz.pl';

    /**
     * @type string
     */
    const SANDBOX_PAYMENTS_HOST = 'https://pay-api.ckpl.us';

    /**
     * @type string
     */
    const SANDBOX_OIDC_HOST = 'https://oidc.ckpl.us';

    /**
     * ConotoxiaPay constructor.
     */
    public function __construct()
    {
        $this->name = 'conotoxiapay';
        $this->tab = 'payments_gateways';
        $this->version = '1.14.3';
        $this->ps_versions_compliancy = ['min' => '1.7', 'max' => _PS_VERSION_];
        $this->author = $this->l('Company name');
        $this->is_eu_compatible = 1;
        $this->need_instance = 1;
        $this->bootstrap = true;
        $this->limited_countries = [];

        parent::__construct();

        $this->displayName = $this->l('Conotoxia Pay');
        $this->description = $this->l('Plugin description');
        $this->confirmUninstall = $this->l('Are you sure you want to uninstall?');

        if (!Configuration::get('MYMODULE_NAME')) {
            $this->warning = $this->l('No name provided');
        }
    }

    public function hookPaymentOptions($params)
    {
        if (!$this->active) {
            return;
        }

        $paymentOption = new PaymentOption();

        $paymentOption->setCallToActionText($this->l('Pay by Conotoxia Pay'))
            ->setAdditionalInformation('<span class="conotoxia-marker-class"></span>')
            ->setModuleName($this->name)
            ->setAction($this->context->link->getModuleLink($this->name, 'payment'));

        if (Configuration::get(Definitions::ICON)) {
            $paymentOption->setLogo($this->getLogo());
        }
        return [$paymentOption];
    }

    public function hookDisplayPaymentEU()
    {
        if (!$this->active) {
            return;
        }

        $paymentOption = new PaymentOption();

        $paymentOption->setCallToActionText($this->l('Pay by Conotoxia Pay'))
            ->setAdditionalInformation('<span class="conotoxia-marker-class"></span>')
            ->setModuleName($this->name)
            ->setAction($this->context->link->getModuleLink($this->name, 'payment'));

        $paymentOption->setLogo($this->getLogo());

        return $paymentOption;
    }

    /**
     * @param string $html
     * @return array
     */
    private function handleGenerateKeys(&$html)
    {
        if (!extension_loaded('openssl')) {
            $html .= $this->displayWarning($this->l('Missing OpenSSL extension.'));
            PrestaShopLogger::addLog('ConotoxiaPay - missing OpenSSL extension', 3, null, 'Pay', null, true);
            return ['public' => null, 'private' => null];
        }

        $keyPair = openssl_pkey_new([
            'private_key_bits' => 2048,
            'private_key_type' => OPENSSL_KEYTYPE_RSA
        ]);
        openssl_pkey_export($keyPair, $privateKey);
        $publicKey = openssl_pkey_get_details($keyPair)['key'];
        openssl_pkey_free($keyPair);
        $html .= $this->displayInformation($this->l('Remember to save configuration after generating keys.'));
        return ['public' => trim($publicKey), 'private' => trim($privateKey)];
    }

    /**
     * @param string $privateKey
     * @return string|null
     */
    private function getPublicKey($privateKey)
    {
        if (!extension_loaded('openssl')) {
            PrestaShopLogger::addLog('ConotoxiaPay - missing OpenSSL extension', 3, null, 'Pay', null, true);
            return null;
        }
        $keyPair = openssl_pkey_get_private($privateKey);
        if ($keyPair === false) {
            return null;
        }
        $publicKey = openssl_pkey_get_details($keyPair)['key'];
        openssl_pkey_free($keyPair);
        return trim($publicKey);
    }

    /**
     * Module configuration form
     *
     * @return string
     * @throws PrestaShopException
     */
    public function getContent()
    {
        $html = '';
        $publicKey = null;
        $privateKey = null;

        if (Tools::getValue('generate_keys') === $_SESSION['generate_keys']) {
            $keys = $this->handleGenerateKeys($html);
            $publicKey = $keys['public'];
            $privateKey = $keys['private'];
        }
        if (Tools::isSubmit('btnSubmit')) {
            $privateKey = trim(Tools::getValue(Definitions::PRIVATE_KEY));
            $publicKey = $this->getPublicKey($privateKey);
            Configuration::updateValue(Definitions::CLIENT_ID, trim(Tools::getValue(Definitions::CLIENT_ID)));
            Configuration::updateValue(Definitions::CLIENT_SECRET, trim(Tools::getValue(Definitions::CLIENT_SECRET)));
            Configuration::updateValue(Definitions::POINT_OF_SALE, trim(Tools::getValue(Definitions::POINT_OF_SALE)));
            Configuration::updateValue(Definitions::PUBLIC_KEY, $publicKey);
            Configuration::updateValue(Definitions::PRIVATE_KEY, $privateKey);
            Configuration::updateValue(Definitions::SANDBOX, Tools::getValue(Definitions::SANDBOX));
            Configuration::updateValue(Definitions::ICON, Tools::getValue(Definitions::ICON));

            if ($this->isConfigurationCompleted()) {
                try {
                    $conPay = $this->initializeConotoxiaPay();
                    $conPay->pickSignatureKey();
                    $conPay->merchant()->sendPublicKey();
                    $html .= $this->displayConfirmation($this->l('Saved public key in Conotoxia Pay.'));
                } catch (CKPL\Pay\Exception\Api\PublicKeyAlreadyExistException $e) {
                    $conPay->merchant()->setPublicKeyId($e->getKid());
                    $html .= $this->displayConfirmation($this->l('Settings saved.'));
                } catch (Exception $e) {
                    PrestaShopLogger::addLog('ConotoxiaPay - saving configuration problem: ' . $e->getMessage(), 2, $e->getCode(), 'Pay', null, true);
                    $html .= $this->displayWarning($this->getSettingsProblemMessage());
                }
            } else {
                if (!empty($privateKey) && $publicKey === null) {
                    $html .= $this->displayWarning($this->l('Wrong private key.'));
                } else {
                    $html .= $this->displayInformation($this->l('Settings saved but some of the fields are missing.'));
                }
            }
        }

        $generationToken = (string)mt_rand();
        $_SESSION['generate_keys'] = $generationToken;

        $fields_form = array(
            'form' => array(
                'legend' => array(
                    'title' => $this->l('Conotoxia Pay configurator'),
                    'icon' => 'icon-envelope'
                ),
                'input' => array(
                    array(
                        'type' => 'text',
                        'label' => $this->l('Client id'),
                        'name' => Definitions::CLIENT_ID,
                        'required' => true
                    ),
                    array(
                        'type' => 'text',
                        'label' => $this->l('Client secret'),
                        'name' => Definitions::CLIENT_SECRET,
                        'required' => true
                    ),
                    array(
                        'type' => 'text',
                        'label' => $this->l('Point of sale'),
                        'name' => Definitions::POINT_OF_SALE,
                        'required' => true
                    ),
                    array(
                        'type' => 'textarea',
                        'label' => $this->l('RSA private key'),
                        'name' => Definitions::PRIVATE_KEY,
                        'required' => true,
                        'hint' => $this->l('RSA private key that will be used to sign request.'),
                        'desc' => nl2br(
                            $this->generateModuleConfigurationLink(['generate_keys' => $generationToken], $this->l('Generate public and private key'))
                        ),
                    ),
                    array(
                        'type' => 'textarea',
                        'label' => $this->l('RSA public key'),
                        'name' => Definitions::PUBLIC_KEY,
                        'readonly' => true,
                        'hint' => nl2br(
                            $this->l('RSA public key that will be sent to Conotoxia Pay.')
                        ),
                    ),
                    array(
                        'type' => 'switch',
                        'label' => $this->l('Sandbox mode'),
                        'name' => Definitions::SANDBOX,
                        'required' => true,
                        'values' => [[
                            'id' => 'active_on',
                            'value' => 1,
                            'label' => $this->l('Enabled')
                        ], [
                            'id' => 'active_off',
                            'value' => 0,
                            'label' => $this->l('Disabled')
                        ]
                        ],
                    ),
                    array(
                        'type' => 'switch',
                        'label' => $this->l('Payment icon'),
                        'name' => Definitions::ICON,
                        'required' => false,
                        'values' => [[
                            'id' => 'active_on',
                            'value' => 1,
                            'label' => $this->l('Enabled')
                        ], [
                            'id' => 'active_off',
                            'value' => 0,
                            'label' => $this->l('Disabled')
                        ]
                        ],
                    ),
                ),
                'submit' => [
                    'title' => $this->l('Save'),
                ]
            ),
        );

        $helper = new HelperForm();
        $helper->name_controller = $this->name;
        $helper->token = Tools::getAdminTokenLite('AdminModules');
        $helper->show_toolbar = false;
        $helper->title = $this->displayName;
        $helper->submit_action = 'btnSubmit';
        $helper->tpl_vars = ['fields_value' => $this->getConfigFieldsValues($privateKey, $publicKey)];

        return $html . $helper->generateForm([$fields_form]);
    }

    public function getConfigFieldsValues($generatedPrivateKey = null, $generatedPublicKey = null)
    {
        return array(
            Definitions::CLIENT_ID => Tools::getValue(Definitions::CLIENT_ID, Configuration::get(Definitions::CLIENT_ID)),
            Definitions::CLIENT_SECRET => Tools::getValue(Definitions::CLIENT_SECRET, Configuration::get(Definitions::CLIENT_SECRET)),
            Definitions::POINT_OF_SALE => Tools::getValue(Definitions::POINT_OF_SALE, Configuration::get(Definitions::POINT_OF_SALE)),
            Definitions::PUBLIC_KEY => $generatedPublicKey ?? Configuration::get(Definitions::PUBLIC_KEY),
            Definitions::PRIVATE_KEY => $generatedPrivateKey ?? Configuration::get(Definitions::PRIVATE_KEY),
            Definitions::SANDBOX => Tools::getValue(Definitions::SANDBOX, Configuration::get(Definitions::SANDBOX)),
            Definitions::ICON => Tools::getValue(Definitions::ICON, Configuration::get(Definitions::ICON))
        );
    }

    public function getLogo($file = 'pay_logo.svg')
    {
        return Media::getMediaPath(_PS_MODULE_DIR_ . $this->name . '/views/img/' . $file);
    }

    public function install()
    {
        if (!parent::install()) {
            return false;
        }
        $installer = InstallerFactory::create();

        return $installer->install($this);
    }

    public function uninstall()
    {
        $installer = InstallerFactory::create();

        return $installer->uninstall() && parent::uninstall();
    }

    /**
     * @return ConfigurationInterface
     * @throws \CKPL\Pay\Exception\ConfigurationException
     */
    protected function makeConfiguration()
    {
        return ConfigurationFactory::fromArray([
            ConfigurationInterface::HOST => $this->getPaymentHost(),
            ConfigurationInterface::OIDC => $this->getOidcHost(),
            ConfigurationInterface::CLIENT_ID => Configuration::get(Definitions::CLIENT_ID),
            ConfigurationInterface::CLIENT_SECRET => Configuration::get(Definitions::CLIENT_SECRET),
            ConfigurationInterface::POINT_OF_SALE => Configuration::get(Definitions::POINT_OF_SALE),
            ConfigurationInterface::PRIVATE_KEY => Configuration::get(Definitions::PRIVATE_KEY),
            ConfigurationInterface::PUBLIC_KEY => Configuration::get(Definitions::PUBLIC_KEY),
            ConfigurationInterface::STORAGE => new PS_Gateway_Conotoxia_Pay_Storage()
        ]);
    }

    private function getPaymentHost()
    {
        if (Configuration::get(Definitions::SANDBOX)) {
            return static::SANDBOX_PAYMENTS_HOST;
        }
        return static::PAYMENTS_HOST;
    }

    private function getOidcHost()
    {
        if (Configuration::get(Definitions::SANDBOX)) {
            return static::SANDBOX_OIDC_HOST;
        }
        return static::OIDC_HOST;
    }

    /**
     * @return Pay
     * @throws Exception
     */
    public function initializeConotoxiaPay()
    {
        try {
            return new Pay($this->makeConfiguration());
        } catch (Exception $e) {
            PrestaShopLogger::addLog('ConotoxiaPay - initialization problem: ' . $e->getMessage(), 3, $e->getCode(), 'Pay', null, true);
            throw $e;
        }
    }

    /**
     * @return bool
     */
    public function isConfigurationCompleted()
    {
        $required = [
            Definitions::CLIENT_ID,
            Definitions::CLIENT_SECRET,
            Definitions::POINT_OF_SALE,
            Definitions::PUBLIC_KEY,
            Definitions::PRIVATE_KEY
        ];
        $result = true;

        foreach ($required as $option) {
            if (empty(Configuration::get($option))) {
                $result = false;
                break;
            }
        }

        return $result;
    }

    public function hookDisplayAdminOrderLeft()
    {
        $vieworder = Tools::getValue('vieworder');
        $id_order = Tools::getValue('id_order');

        $show_refunds = false;
        $refunds = [];
        if ($vieworder !== false && $id_order !== false) {
            $order = new Order($id_order);

            if ($order->module === 'conotoxiapay') {
                $refunds = RefundRepository::getRefundsByOrderId($id_order);
                if (!empty($refunds)) {
                    $show_refunds = true;
                }
            }
        }

        $this->context->smarty->assign('show_refunds', $show_refunds);
        $this->context->smarty->assign('refunds', $refunds);

        return $this->display(__FILE__, '/views/templates/admin/refunds.tpl');
    }

    public function hookBackOfficeHeader()
    {
        $viewOrder = Tools::getValue('vieworder');
        $orderId = Tools::getValue('id_order');

        $showRefundComponent = false;
        if ($viewOrder !== false && $orderId !== false) {
            $order = new Order($orderId);

            if ($order->module === 'conotoxiapay') {
                $orderPayment = $order->getOrderPayments();
                $paymentId = $orderPayment[0]->transaction_id;

                if (PaymentRepository::isBooked($paymentId) && !RefundRepository::isAnyFullRefund($paymentId)) {
                    $showRefundComponent = true;
                    $this->context->smarty->assign('conpay_payment_id', $paymentId);
                }
            }
        }

        $this->context->smarty->assign('conpay_show_refund', $showRefundComponent);
        if (!$showRefundComponent) {
            return $this->display(__FILE__, '/views/templates/admin/header.tpl');
        }

        $showSuccess = false;
        $showError = false;
        $alertMessage = '';
        $locale = $this->context->getCurrentLocale();

        $refundType = Tools::getValue('conpay_refund_type');
        $refundAmount = str_replace(',', '.', $refundType === 'FULL' ? $order->total_paid : Tools::getValue('conpay_refund_amount'));
        $reason = empty(Tools::getValue('conpay_refund_reason')) ? $this->l('Refund reason') : Tools::getValue('conpay_refund_reason');

        $this->context->smarty->assign('conpay_refund_type', $refundType);
        $this->context->smarty->assign('conpay_refund_amount', is_numeric($refundAmount) ? $locale->formatNumber($refundAmount) : $refundAmount);
        $this->context->smarty->assign('conpay_refund_full_amount', $locale->formatNumber($order->total_paid));
        $this->context->smarty->assign('conpay_refund_reason', $reason);

        if ($_SERVER['REQUEST_METHOD'] === 'POST' && Tools::getValue('conpay_refund_submit') && Tools::getValue('conpay_token_check') === $_SESSION['conpay_form_token']) {
            if (!is_numeric($refundAmount)) {
                $showError = true;
                $alertMessage = $this->l('Wrong refund amount');
            } else if ($refundAmount > $order->total_paid) {
                $showError = true;
                $alertMessage = $this->l('The refund amount you entered is greater than paid amount');
            } else {
                try {
                    $refundId = $this->conOrderRefund($refundAmount, $order, $reason)->getId();
                    RefundRepository::createRefund($orderId, $refundId, $paymentId, $refundType, $refundAmount, $order->id_currency);

                    $history = new OrderHistory();
                    $history->id_order = (int)$orderId;
                    $history->id_employee = (int)$this->context->employee->id;
                    $history->changeIdOrderState(Configuration::get('PS_OS_REFUND'), $orderId);
                    $history->addWithemail(true, array());

                    $showSuccess = true;
                    $alertMessage = sprintf($this->l('Successfully created refund'), $refundId);
                } catch (CKPL\Pay\Exception\Http\HttpConflictException | CKPL\Pay\Exception\Http\HttpNotFoundException $e) {
                    $showError = true;
                    $alertMessage = $e->getTranslatedMessage($this->context->language->iso_code) ?? $this->getRefundCreationProblemMessage();
                } catch (Exception $e) {
                    $showError = true;
                    $alertMessage = $this->getRefundCreationProblemMessage();
                }
            }
        }

        $formToken = (string)mt_rand();
        $_SESSION['conpay_form_token'] = $formToken;
        $this->context->smarty->assign('conpay_form_token', $formToken);

        $this->context->smarty->assign('show_success', $showSuccess);
        $this->context->smarty->assign('show_error', $showError);
        $this->context->smarty->assign('alert_message', $alertMessage);

        return $this->display(__FILE__, '/views/templates/admin/header.tpl');
    }

    /**
     * @return string
     * @throws PrestaShopException
     */
    private function getSettingsProblemMessage()
    {
        return sprintf($this->l('Settings saved with errors.'), $this->generateLogsLink($this->l('Logs')));
    }

    /**
     * @return string
     * @throws PrestaShopException
     */
    private function getRefundCreationProblemMessage()
    {
        return sprintf($this->l('Refund creation problem'), $this->generateLogsLink($this->l('Logs')));
    }

    /**
     * @param array $queryParameters
     * @param string $linkText
     * @return string
     * @throws PrestaShopException
     */
    private function generateModuleConfigurationLink($queryParameters, $linkText)
    {
        return $this->generateLink($this->context->link->getAdminLink(
            'AdminModules',
            true,
            [],
            array_merge(['configure' => 'conotoxiapay'], $queryParameters)
        ), $linkText);
    }

    /**
     * @param $linkText string
     * @return string
     * @throws PrestaShopException
     */
    private function generateLogsLink($linkText)
    {
        return $this->generateLink($this->context->link->getAdminLink('AdminLogs', true), $linkText);
    }

    /**
     * @param string $address
     * @param string $label
     * @return string
     */
    private function generateLink($address, $label)
    {
        return '<a href="' . $address . '">' . $label . '</a>';
    }

    /**
     * @param float $amount
     * @param Order $order
     *
     * @param string $reason
     * @return \CKPL\Pay\Model\Response\CreatedRefundResponseModel|string
     *
     * @throws \CKPL\Pay\Exception\ClientException
     * @throws \CKPL\Pay\Exception\Exception
     */
    private function conOrderRefund($amount, $order, $reason)
    {
        try {
            $orderPayment = $order->getOrderPayments();

            $currency = Currency::getCurrency($order->id_currency);
            $conPay = $this->initializeConotoxiaPay();
            $refundBuilder = $conPay->refunds()->createRefundBuilder();
            $refund = $refundBuilder
                ->setPaymentId($orderPayment[0]->transaction_id)
                ->setExternalRefundId((string)$order->id)
                ->setReason($reason)
                ->setValue((string)$amount)
                ->setCurrency($currency['iso_code'])
                ->getRefund();
            return $conPay->refunds()->makeRefund($refund);
        } catch (CKPL\Pay\Exception\ClientException | \CKPL\Pay\Exception\Exception $e) {
            PrestaShopLogger::addLog('ConotoxiaPay - refund creation problem: ' . $e->getMessage(), 3, $e->getCode(), 'Order', $order->id, true);
            throw $e;
        }
    }

    public function buildTemplatePath($name)
    {
        return 'module:conotoxiapay/views/templates/front/' . $name . '.tpl';
    }
}
